import sys

if sys.stdout.isatty():
    cbRed = "\x1b[1;31m"
    cbGreen = "\x1b[1;32m"
    cbYellow = "\x1b[1;33m"
    cReset = "\x1b[0m"
else:
    cbRed = ""
    cbGreen = ""
    cbYellow = ""
    cReset = ""

the_output = None


class Output:
    def __init__(self, args):
        global the_output
        self._verbosity = self._verbosity(args)
        the_output = self

    def verbosity(self):
        return self._verbosity

    def output(self, level: int, msg: str):
        if self._verbosity > level:
            print(msg)

    def error(self, title: str, msg: str):
        self.output(0, f"{cbRed}{title}{cReset}: {msg}")

    def warn(self, title: str, msg: str):
        self.output(1, f"{cbYellow}{title}{cReset}: {msg}")

    def success(self, title: str, msg: str):
        self.output(0, f"{cbGreen}{title}{cReset}: {msg}")

    def expected_failure(self, title: str, msg: str):
        self.output(0, f"{cbRed}{title}{cReset}: {msg}")

    def _verbosity(self, args):
        if args.quiet:
            return 0

        if args.trace:
            return 100

        if args.verbose:
            return 2

        return 1


def global_output() -> Output:
    return the_output



from importlib.resources import files

OP_ERASE = "erase"
OP_CHECK = "check"
ATF_1502 = "ATF1502"
ATF_1504 = "ATF1504"
ATF_1508 = "ATF1508"

STANDARD_VECTORS = {
    OP_ERASE: {
        ATF_1502: files("atfu.standard_vectors").joinpath("e1502.xsvf"),
        ATF_1504: files("atfu.standard_vectors").joinpath("e1504.xsvf"),
        ATF_1508: files("atfu.standard_vectors").joinpath("e1508.xsvf"),
    },
    OP_CHECK: {
        ATF_1502: files("atfu.standard_vectors").joinpath("bc1502.xsvf"),
        ATF_1504: files("atfu.standard_vectors").joinpath("bc1504.xsvf"),
        ATF_1508: files("atfu.standard_vectors").joinpath("bc1508.xsvf"),
    },
}


def find_vector_file(operation: str, device_key: str):
    return STANDARD_VECTORS.get(operation, {}).get(device_key)


def atf_device_key(number: int):
    if number == 1502:
        return ATF_1502
    elif number == 1504:
        return ATF_1504
    elif number == 1508:
        return ATF_1508
    else:
        raise ValueError(f"{number} is not valid supported ATF CPLD device")



import serial.tools.list_ports
from serial.serialutil import SerialException
from tabulate import tabulate

OUR_VID = 0x04D8
OUR_PID = 0xE5C7


def handle_list(args):
    if args.plain:
        _list_plain(find_programmers())
    else:
        _list_detail(find_programmers())


def _list_plain(programmers: list[list[str]]) -> None:
    for programmer in programmers:
        print(programmer["Device"])


def _list_detail(programmers: list[list[str]]) -> None:
    print()
    if programmers:
        print(f"Found {len(programmers)} programmer(s):")
        print()
        print(tabulate(programmers, headers="keys", showindex=True))
        print()
        print(
            "Unless specified (with the -p option), ",
            f"{programmers[0]['Device']} will be used",
        )
        print("for programming and related operations.")
    else:
        print("No programmers found!")
    print()


def handle_query(args):
    print()
    print(f"Querying programmer '{args.programmer}'")

    try:
        port = serial.Serial(port=args.programmer, timeout=5)
    except SerialException as e:
        print(f"  => Unable to open programmer ({e})")
        exit(2)

    port.timeout = 2
    port.flushInput()
    port.flushOutput()
    port.write(b"V")

    ver1 = port.readline()
    ver2 = port.readline()

    if not ver1 or not ver2:
        print("  => Timed out")

    if ver1 != b"\r\n":
        print("  => Malformed response")
        exit(1)

    verfields = ver2.decode("ascii").split(",")
    if verfields[0] != "Q0":
        print("  => Malformed response")
        exit(1)

    print(f"  => Firmware identifies as: {verfields[1].strip()}")
    print()
    exit(0)


def find_programmers() -> list[dict[str, str]]:
    all_progs = serial.tools.list_ports.comports(False)
    return [
        {
            "Device": prog.device,
            "Product": prog.product,
            "Manufacturer": prog.manufacturer,
        }
        for prog in all_progs
        if prog.vid == OUR_VID and prog.pid == OUR_PID
    ]


def default_programmer() -> dict[str, str]:
    default = find_programmers()[0:1]
    if default:
        return default[0]
    else:
        return None


def default_programmer_path() -> str:
    default = default_programmer()
    if default:
        return default["Device"]
    else:
        return None



import argparse

import atfu.program
import atfu.erase
import atfu.check
import atfu.programmer

BOARDNAME = "Little ATF150x Programmer Board"
PROGNAME = f"{BOARDNAME} Utility"
VERSION = "2024.1014"


def main():
    args = _arg_parser().parse_args()
    args.func(args)


def _arg_parser():
    parser = argparse.ArgumentParser(description=f"{PROGNAME}")
    parser.add_argument(
        "--version",
        action="version",
        version=f"{PROGNAME} v{VERSION}; (c) The Really Old-School Company Ltd",
    )
    parser.add_argument(
        "-q", "--quiet", action="store_true", help="Silence almost all output"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Allow additional output"
    )
    parser.add_argument(
        "-t",
        "--trace",
        action="store_true",
        help="Enable debugging output (can be noisy!)",
    )

    subs = parser.add_subparsers(dest="{program,erase,check,programmer}", required=True)

    # Program
    program = subs.add_parser("program", help="Program an ATF150x device")
    program.add_argument(
        "-e", "--erase", action="store_true", help="Erase before programming"
    )
    program.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Force-erase before programming (implies -e)",
    )
    program.add_argument(
        "-d",
        "--device",
        default="ATF1502",
        choices=("ATF1502", "ATF1504", "ATF1508"),
        help="Device to program (default: %(default)s)",
    )
    program.add_argument(
        "-p",
        "--programmer",
        default=atfu.programmer.default_programmer_path(),
        help="Programmer device (default: %(default)s)",
    )
    program.add_argument(
        "filename",
        nargs="+",
        type=argparse.FileType("rb"),
        help=".jed, .svf or .xsvf file(s) to program",
    )
    program.set_defaults(func=atfu.program.handler)

    # Erase
    erase = subs.add_parser("erase", help="Erase an ATF150x device")
    erase.add_argument("-f", "--force", action="store_true", help="Force erase")
    erase.add_argument(
        "-p",
        "--programmer",
        default=atfu.programmer.default_programmer_path(),
        help="Programmer device (default: %(default)s)",
    )
    erase.add_argument(
        "-d",
        "--device",
        default="ATF1502",
        choices=("ATF1502", "ATF1504", "ATF1508"),
        help="Device to erase (default: %(default)s)",
    )
    erase.set_defaults(func=atfu.erase.handler)

    # Check
    check = subs.add_parser("check", help="Check if an ATF150x device is blank")
    check.add_argument(
        "-p",
        "--programmer",
        default=atfu.programmer.default_programmer_path(),
        help="Programmer device (default: %(default)s)",
    )
    check.add_argument(
        "-d",
        "--device",
        default="ATF1502",
        choices=("ATF1502", "ATF1504", "ATF1508"),
        help="Device to check (default: %(default)s)",
    )
    check.set_defaults(func=atfu.check.handler)

    # Programmer
    programmer = subs.add_parser("programmer", help=f"{BOARDNAME} device functions")

    programmer_subs = programmer.add_subparsers(required=True)

    programmer_list = programmer_subs.add_parser(
        "list", description="Functions for listing connected programmers"
    )
    programmer_list.add_argument(
        "--plain", action="store_true", help="Display a plain list of device paths"
    )
    programmer_list.set_defaults(func=atfu.programmer.handle_list)

    programmer_query = programmer_subs.add_parser(
        "query", description="Functions for querying connected programmers"
    )
    programmer_query.add_argument(
        "-p",
        "--programmer",
        default=atfu.programmer.default_programmer_path(),
        help="Programmer device to query (default: %(default)s)",
    )
    programmer_query.set_defaults(func=atfu.programmer.handle_query)

    return parser


if __name__ == "__main__":
    main()



from atfu.little_board.jtag_programmer import JtagProgrammer
from atfu.standard_vectors import OP_CHECK, find_vector_file
from atfu.output import Output


def handler(args):
    exit(perform_check(args))


def perform_check(args) -> int:
    output = Output(args)

    vector_path = find_vector_file(OP_CHECK, args.device)

    if vector_path is None:
        output.error("Failure", f"{args.device} is not a recognised device.")
        return 10

    try:
        vector_fd = vector_path.open("rb")
    except FileNotFoundError:
        output.error(
            "Failure", f"Support for {args.device} is not installed correctly."
        )
        output.error(
            "       ", "Please check your installation and reinstall if necessary."
        )
        return 2

    prog = JtagProgrammer(
        "CHECK",
        args.programmer,
        output.verbosity(),
        no_filename=True,
        no_success=True,
        no_fail=True,
    )

    if prog.upload_one_file(vector_fd):
        output.success(args.device, "is blank")
        return 0
    else:
        output.expected_failure(args.device, "is not blank")
        return 1



# Vendored from glasgow.protocol.jesd3

# Ref: JEDEC JESD3-C
# Accession: G00029

import re
from bitarray import bitarray
from atfu.output import global_output

__all__ = ["JESD3Parser", "JESD3ParsingError"]


class JESD3ParsingError(Exception):
    pass


class JESD3Lexer:
    """
    A JESD3 (JED) lexer.

    :type buffer: str
    :attr buffer:
        Input buffer.

    :type position: int
    :attr position:
        Offset into buffer from which the next field will be read.
    """

    # This follows the JESD3-C grammar, with the exception that spaces are more permissive.
    # As described, only 0x0D is allowed in between fields, which is absurd.
    _fields = (
        (r"N", r"[ \r\n]*(.*?)"),
        (r"D", r".*?"),
        (r"QF", r"([0-9]+)"),
        (r"QP", r"([0-9]+)"),
        (r"QV", r"([0-9]+)"),
        (r"F", r"([01])"),
        (r"L", r"([0-9]+)[ \r\n]+([01 \r\n]+)"),
        (r"C", r"([0-9A-F]{4})"),
        (r"EH", r"([0-9A-F]+)"),
        (r"E", r"([01]+)"),
        (r"UA", r"([\r\n\x20-\x29\x2B-\x7E]+)"),
        (r"UH", r"([0-9A-F]+)"),
        (r"U", r"([01]+)"),
        (r"J", r"([0-9]+)[ \r\n]+([0-9]+)"),
        (r"G", r"([01])"),
        (r"X", r"([01])"),
        (r"P", r"([ \r\n]*[0-9]+)+"),
        # This next one _can_ handle the QV (and has all vector types) but still won't checksum
        #
        # TODO fix this properly (currently we just error out with a "helpful" message)
        #        (r"V", r"([0-9]+)[ \r\n]+(?:QV\d+\* )?([0-9BCDFHLNPTUXZ]+)"),
        (r"V", r"([0-9]+)[ \r\n]+([0-9BCDFHLNPTUXZ]+)"),
        (r"S", r"([01]+)"),
        (r"R", r"([0-9A-F]{8})"),
        (r"T", r"([0-9]+)"),
        (r"A", r"([\r\n\x20-\x29\x2B-\x7E]*)([0-9]+)"),
    )
    _stx_spec_re = re.compile(r"\x02(.*?)\*[ \r\n]*", re.A | re.S)
    _stx_quirk_re = re.compile(r"\x02()[ \r\n]*", re.A | re.S)
    _etx_re = re.compile(r"\x03([0-9A-F]{4})", re.A | re.S)
    _ident_re = re.compile(r"|".join(ident for ident, args in _fields), re.A | re.S)
    _field_res = {
        ident: re.compile(ident + args + r"[ \r\n]*\*[ \r\n]*", re.A | re.S)
        for ident, args in _fields
    }

    def __init__(self, buffer, quirk_no_design_spec=False):
        self.buffer = buffer
        self.position = 0
        self.checksum = 0
        self._state = "start"
        if quirk_no_design_spec:
            self._stx_re = self._stx_quirk_re
        else:
            self._stx_re = self._stx_spec_re

    def line_column(self, position=None):
        """
        Return a ``(line, column)`` tuple for the given or, if not specified, current position.

        Both the line and the column start at 1.
        """
        line = len(re.compile(r"\n").findall(self.buffer, endpos=self.position))
        if line > 1:
            column = self.position - self.buffer.rindex("\n", 0, self.position)
        else:
            column = self.position
        return line + 1, column + 1

    def __iter__(self):
        return self

    def __next__(self):
        """Return the next token and advance the position."""
        if self._state == "start":
            match = self._stx_re.search(self.buffer, self.position)
            if not match:
                raise JESD3ParsingError("could not find STX marker")
            else:
                token = "start"
                self._state = "fields"
                self.checksum += sum(map(ord, match.group(0)))

        elif self._state == "fields":
            match = self._ident_re.match(self.buffer, self.position)

            # Another horrible hack - we can't handle empty fields, so just try to skip them...
            # Will be replaced by new parser, so there's that to look forward to at least...
            #
            if not match and self.buffer[self.position] == "*":
                while self.buffer[self.position] in "*\n ":
                    self.position += 1
                match = self._ident_re.match(self.buffer, self.position)
            #
            #

            if match:
                token = match.group(0)
                match = self._field_res[token].match(self.buffer, self.position)

                if not match:
                    if token == "V":
                        global_output().error(
                            "Failure",
                            "JEDEC file contains test vectors, which cannot be processed by this tool. Please rebuild without them.",
                        )
                        exit(1000)
                    else:
                        raise JESD3ParsingError(
                            "field %s has invalid format at line %d, column %d"
                            % (token, *self.line_column())
                        )
                else:
                    self.checksum += sum(map(ord, match.group(0)))

            else:
                match = self._etx_re.match(self.buffer, self.position)
                if not match:
                    raise JESD3ParsingError(
                        "unrecognized field at line %d, column %d (%r...)"
                        % (
                            *self.line_column(),
                            self.buffer[self.position : self.position + 16],
                        )
                    )
                else:
                    token = "end"
                    self._state = "end"
                    self.checksum += 0x03

        elif self._state == "end":
            raise StopIteration

        self.position = match.end()
        return token, match.start(), match.groups()


class JESD3Parser:
    def __init__(self, buffer, **kwargs):
        self._lexer = JESD3Lexer(buffer, **kwargs)
        self._position = 0

        self.design_spec = ""
        self.notes = []
        self.fuse = None
        self._fuse_default = None
        self._fuse_bit_count = 0
        self.electrical_fuse = None
        self.user_fuse = None
        self.security_fuse = None
        self.device_id = None

    def _parse_error(self, error):
        raise JESD3ParsingError(
            "%s at line %d, column %d"
            % (error, *self._lexer.line_column(self._position))
        )

    def parse(self):
        for token, position, args in self._lexer:
            self._position = position
            # print("lexem: %r %r" % (token, args))
            getattr(self, "_on_" + token)(*args)

    def _on_start(self, design_spec):
        """Start marker and design specification"""
        self.design_spec = design_spec

    def _on_N(self, note):
        """Note"""
        self.notes.append(note)

    def _on_D(self):
        """Device (obsolete)"""

    def _on_QF(self, count):
        """Fuse count"""
        if self.fuse is not None:
            self._parse_error("fuse count specified more than once")
        self.fuse = bitarray(int(count, 10), endian="little")

    def _on_QP(self, count):
        """Pin count (unsupported and ignored)"""

    def _on_QV(self, count):
        """Test vector count (unsupported)"""
        if int(count, 10) > 0:
            self._parse_error("test vectors are unsupported")

    def _on_F(self, state):
        """Fuse default state"""
        if self.fuse is None:
            self._parse_error("fuse default state specified before fuse count")
        if self._fuse_default is not None:
            self._parse_error("fuse default state specified more than once")
        if self._fuse_bit_count > 0:
            self._parse_error("fuse default state specified after fuse list")
        self._fuse_default = int(state, 2)
        self.fuse.setall(self._fuse_default)

    def _on_L(self, index, values):
        """Fuse list"""
        if self.fuse is None:
            self._parse_error("fuse list specified before fuse count")
        index = int(index, 10)
        values = bitarray(re.sub(r"[ \r\n]", "", values), endian="little")
        if index + len(values) > len(self.fuse):
            self._parse_error(
                "fuse list specifies range [%d:%d] beyond last fuse %d"
                % (index, index + len(values), len(self.fuse))
            )
        self.fuse[index : index + len(values)] = values
        self._fuse_bit_count += len(values)

    def _on_C(self, checksum):
        """Fuse checksum"""
        expected_checksum = int(checksum, 16)
        actual_checksum = sum(self.fuse.tobytes()) & 0xFFFF
        if expected_checksum != actual_checksum:
            self._parse_error(
                "fuse checksum mismatch: expected %04X, actual %04X"
                % (expected_checksum, actual_checksum)
            )

    def _set_electrical_fuse(self, value):
        if self.electrical_fuse is not None:
            self._parse_error("electrical fuse specified more than once")
        self.electrical_fuse = value

    def _on_EH(self, value):
        """Electrical fuse, hex"""
        self._set_electrical_fuse(int(value, 16))

    def _on_E(self, value):
        """Electrical fuse, binary"""
        self._set_electrical_fuse(int(value, 2))

    def _set_user_fuse(self, value):
        if self.user_fuse is not None:
            self._parse_error("user fuse specified more than once")
        self.user_fuse = value

    def _on_UA(self, value):
        """User fuse, 7-bit ASCII"""
        int_value = 0
        for char in reversed(value):
            int_value <<= 7
            int_value |= ord(char)
        self._set_user_fuse(int_value)

    def _on_UH(self, value):
        """User fuse, hex"""
        self._set_user_fuse(int(value, 16))

    def _on_U(self, value):
        """User fuse, binary"""
        self._set_user_fuse(int(value, 2))

    def _on_J(self, arch_code, pinout_code):
        """Device identification"""
        if self.device_id is not None:
            self._parse_error("device identification specified more than once")
        self.device_id = (int(arch_code, 10), int(pinout_code, 10))

    def _on_G(self, value):
        """Security fuse"""
        if self.security_fuse is not None:
            self._parse_error("security fuse specified more than once")
        self.security_fuse = int(value, 2)

    def _on_X(self, value):
        """Default test condition (unsupported and ignored)"""

    def _on_P(self, pin_numbers):
        """Pin list (unsupported and ignored)"""

    def _on_V(self, vector_number, test_conditions):
        """Test vector (unsupported and ignored)"""

    def _on_S(self, test_condition):
        """Signature analysis starting vector (unsupported)"""
        self._parse_error("signature analysis is not supported")

    def _on_R(self, test_sum):
        """Signature analysis resulting vector (unsupported and ignored)"""

    def _on_T(self, test_cycles):
        """Signature analysis test cycle count (unsupported and ignored)"""

    def _on_A(self, subfield, delay):
        """Propagation delay for test vectors (unsupported and ignored)"""

    def _on_end(self, checksum):
        """End marker and checksum"""
        expected_checksum = int(checksum, 16)
        if expected_checksum == 0x0000:
            return
        actual_checksum = self._lexer.checksum & 0xFFFF
        if expected_checksum != actual_checksum:
            self._parse_error(
                "transmission checksum mismatch: expected %04X, actual %04X"
                % (expected_checksum, actual_checksum)
            )

        if self._fuse_default is None and self._fuse_bit_count < len(self.fuse):
            self._parse_error(
                "fuse default state is not specified, and only %d out of %d fuse "
                "bits are explicitly defined" % (self._fuse_bit_count, len(self.fuse))
            )


if __name__ == "__main__":
    import sys

    with open(sys.argv[1], "r") as f:
        parser = JESD3Parser(f.read(), quirk_no_design_spec=False)
        parser.parse()
        for i in range(0, len(parser.fuse) + 63, 64):
            print("%08x: %s" % (i, parser.fuse[i : i + 64].to01()))



from atfu.little_board.jtag_programmer import JtagProgrammer
from atfu.standard_vectors import OP_ERASE, find_vector_file
from atfu.output import Output


def handler(args):
    exit(perform_erase(args))


def perform_erase(args, no_success=True) -> int:
    output = Output(args)
    vector_path = find_vector_file(OP_ERASE, args.device)

    if vector_path is None:
        output.error("Failure", f"{args.device} is not a recognised device.")
        return 10

    try:
        vector_fd = vector_path.open("rb")
    except FileNotFoundError:
        output.error(
            "Failure", f"Support for {args.device} is not installed correctly."
        )
        output.error(
            "       ", "Please check your installation and reinstall if necessary."
        )
        return 2

    if args.force:
        prog = JtagProgrammer(
            "ERASE-1",
            args.programmer,
            output.verbosity(),
            no_filename=True,
            no_fail=True,
            no_success=no_success,
        )
        ok = prog.upload_one_file(vector_fd, JtagProgrammer.MODE_HIVPP)
        if not ok:
            prog = JtagProgrammer(
                "ERASE-2",
                args.programmer,
                output.verbosity(),
                no_filename=True,
                no_success=no_success,
            )
            if not prog.upload_one_file(vector_fd, JtagProgrammer.MODE_NORMAL):
                return 1
            else:
                return 0
        else:
            return 0
    else:
        prog = JtagProgrammer(
            "ERASE",
            args.programmer,
            output.verbosity(),
            no_filename=True,
            no_success=no_success,
        )
        if prog.upload_one_file(vector_fd, JtagProgrammer.MODE_NORMAL):
            return 0
        else:
            return 1



# Vendored from glasgow.protocol.jtag_svf

# Ref: https://www.asset-intertech.com/eresources/svf-serial-vector-format-specification-jtag-boundary-scan
# Accession: G00022
# Ref: http://www.jtagtest.com/pdf/svf_specification.pdf
# Accession: G00023

from abc import ABCMeta, abstractmethod

__all__ = ["SVFEventHandler"]


class SVFEventHandler(metaclass=ABCMeta):
    """
    An abstract base class for Serial Vector Format parsing events.

    The methods of this class are called when a well-formed SVF command is encountered.
    The parser takes care of maintaining all lexical state (e.g. "sticky" parameters),
    but all logical state is maintained by the event handler.
    """

    @abstractmethod
    def svf_frequency(self, frequency):
        """Called when the ``FREQUENCY`` command is encountered."""

    @abstractmethod
    def svf_trst(self, mode):
        """Called when the ``TRST`` command is encountered."""

    @abstractmethod
    def svf_state(self, state, path):
        """Called when the ``STATE`` command is encountered."""

    @abstractmethod
    def svf_endir(self, state):
        """Called when the ``ENDIR`` command is encountered."""

    @abstractmethod
    def svf_enddr(self, state):
        """Called when the ``ENDDR`` command is encountered."""

    @abstractmethod
    def svf_hir(self, tdi, smask, tdo, mask):
        """Called when the ``HIR`` command is encountered."""

    @abstractmethod
    def svf_sir(self, tdi, smask, tdo, mask):
        """Called when the ``SIR`` command is encountered."""

    @abstractmethod
    def svf_tir(self, tdi, smask, tdo, mask):
        """Called when the ``TIR`` command is encountered."""

    @abstractmethod
    def svf_hdr(self, tdi, smask, tdo, mask):
        """Called when the ``HDR`` command is encountered."""

    @abstractmethod
    def svf_sdr(self, tdi, smask, tdo, mask):
        """Called when the ``SDR`` command is encountered."""

    @abstractmethod
    def svf_tdr(self, tdi, smask, tdo, mask):
        """Called when the ``TDR`` command is encountered."""

    @abstractmethod
    def svf_runtest(
        self, run_state, run_count, run_clock, min_time, max_time, end_state
    ):
        """Called when the ``RUNTEST`` command is encountered."""

    @abstractmethod
    def svf_piomap(self, mapping):
        """Called when the ``PIOMAP`` command is encountered."""

    @abstractmethod
    def svf_pio(self, vector):
        """Called when the ``PIO`` command is encountered."""


# -------------------------------------------------------------------------------------------------



import os
import string
import time
import serial
import sys

import serial.serialutil

from progress.bar import Bar
from tqdm import tqdm

if sys.stdout.isatty():
    cbWhite = "\x1b[1;37m"
    cbRed = "\x1b[1;31m"
    cbGreen = "\x1b[1;32m"
    cbYellow = "\x1b[1;33m"
    cReset = "\x1b[0m"
else:
    cbWhite = ""
    cbRed = ""
    cbGreen = ""
    cbYellow = ""
    cReset = ""


class JtagProgrammer(object):
    MODE_NORMAL = b"G"
    MODE_HIVPP = b"H"

    _translate_str_1 = "".join(
        [(chr(x) in string.printable) and chr(x) or "." for x in range(256)]
    )
    _translate_str = bytes(_translate_str_1, "ascii")

    def __init__(
        self,
        operation: str,
        device: str,
        verbosity=0,
        no_filename=False,
        no_fail=False,
        no_success=False,
    ):
        self._operation = operation
        self._device = device
        self._serial = None
        self._need_lf = False
        self._file_size = 0
        self._sum = 0
        self._start_time = 0
        self._error_code = 0
        self._verbosity = verbosity
        self._no_filename = no_filename
        self._no_fail = no_fail
        self._no_success = no_success

    def init_programmer(self, mode):
        if self._serial is None:
            self._serial = serial.Serial(port=self._device)

        self._serial.flushInput()
        self._serial.flushOutput()
        self._serial.write(mode)
        self._start_time = 0

    def print_lf(self):
        if self._need_lf:
            self._need_lf = False
            print()

    def initialize_hashes(self):
        self._sum = 0

    def update_hashes(self, s):
        for c in s:
            self._sum += c

    def print_hashes(self, success: bool):
        cksum = (-self._sum) & 0xFF
        if self._verbosity > 1:
            print("  Expected checksum:  0x%02X/%lu." % (cksum, self._file_size))
            print("  Expected sum: 0x%08lX/%lu." % (self._sum, self._file_size))

        if self._start_time > 0 and self._verbosity > 0:
            if success:
                if not self._no_success:
                    print(
                        f"{cbGreen}Success{cReset} in %.02f seconds."
                        % (time.time() - self._start_time)
                    )
            elif not self._no_fail:
                print(
                    f"{cbRed}Failure{cReset} in %.02f seconds."
                    % (time.time() - self._start_time)
                )

    def upload_one_file(self, fd, mode=MODE_NORMAL):
        try:
            self.init_programmer(mode)
        except serial.serialutil.PortNotOpenError:
            if self._verbosity > 0:
                print(
                    f"{cbRed}Failure{cReset}: Device not found, or communication failure"
                )
            return False

        self._file_size = os.fstat(fd.fileno()).st_size
        bytes_written = 0
        okquit = False

        progress = None
        if self._verbosity > 0:
            progress = tqdm(
                total=self._file_size,
                desc=f"{cbWhite}%-10s{cReset}" % self._operation,
                unit="b",
            )

        while True:
            line = self._serial.readline().strip()

            if not line:
                continue

            command = chr(line[0])
            argument = line[1:].decode("ascii")

            if command == "S":
                # Send data
                try:
                    num_bytes = int(
                        argument
                    )  # Blindly try to continue; _should_ only happen
                    # when programmer has trace debug on and is dumping an
                    # XCOMMENT(STATE)
                except Exception:
                    print(
                        f"{cbYellow}WARN{cReset}: Invalid 'S' command from programmer; Skipped..."
                    )
                    continue

                xsvf_data = fd.read(num_bytes)
                bytes_written += len(xsvf_data)
                self.update_hashes(xsvf_data)
                xsvf_data += b"\xFF" * (num_bytes - len(xsvf_data))
                self._serial.write(xsvf_data)
                if self._verbosity > 0:
                    progress.update(n=num_bytes)

            elif command == "R":
                # Ready
                self.initialize_hashes()
                if self._verbosity > 1 and not self._no_filename:
                    print("File: %s" % os.path.realpath(fd.name))
                self._start_time = time.time()

            elif command == "Q":
                # Quit (maybe error)
                if not okquit:
                    okquit = True
                    continue

                if self._verbosity > 0:
                    progress.close()

                # first field is the error code, second is the message.
                args = argument.split(",")
                self._error_code = int(args[0])
                if self._verbosity > 1:
                    print("Quit: {1:s} ({0:d}).".format(self._error_code, args[1]))
                self.print_hashes(self._error_code == 0)
                return self._error_code == 0

            elif command == "D":
                # Debug
                if self._verbosity > 1:
                    self.print_lf()
                    print("Device:", argument)

            elif command == "!":
                # Important
                if self._verbosity > 1:
                    # Can close this here, programmer only importants when done...
                    #
                    # (unless built with trace but that'll only be internal anyhow...)
                    progress.close()

                    self.print_lf()
                    print(f"{cbGreen}>>>{cReset} ", argument)

            else:
                self.print_lf()
                print(
                    "Unknown command from programmer:",
                    line.translate(JtagProgrammer._translate_str),
                )

    def upload_all_files(self, fd_list):
        ok = True
        for fd in fd_list:
            with fd:
                ok = self.upload_one_file(fd)
                if not ok:
                    break
        return ok



import io
import tempfile
from atfu.output import Output
from atfu.converter.fuseconv import read_jed, write_svf
from atfu.converter.device import ATF1502ASDevice, ATF1504ASDevice, ATF1508ASDevice

# TODO don't be exit()ing in here.......


def jed2svf(
    output: Output, device: str, infile: io.BufferedReader
) -> io.BufferedReader:
    output.output(2, "Converting JED to SVF...")

    # Underlying code expects a text reader, not binary...
    inname = infile.name
    infile.close()
    infile = open(inname, "r")

    if device == "ATF1502":
        c_device = ATF1502ASDevice
    elif device == "ATF1504":
        c_device = ATF1504ASDevice
    elif device == "ATF1508":
        c_device = ATF1508ASDevice
    else:
        output.error(
            "Bug",
            f"Incorrect device string {device} passed in jed2svf; Please report: https://github.com/roscopeco/atfprog-tools/issues",
        )
        exit(1000)

    jed_bits = svf_bits = None
    if infile.name.lower().endswith(".jed"):
        jed_bits, comment = read_jed(infile)
        if c_device.fuse_count != len(jed_bits):
            output.error(
                "Failure",
                f"Wrong number of fuses {len(jed_bits)} for device {device} (expected {c_device.fuse_count})",
            )
            exit(50)
    else:
        output.error(
            "Bug",
            "jed2svf called on non-JED; Please report: https://github.com/roscopeco/atfprog-tools/issues",
        )
        exit(1000)

    if svf_bits is None:
        svf_bits = c_device.jed_to_svf(jed_bits)

    outfile = tempfile.NamedTemporaryFile(prefix="atfu-temp-", suffix=".svf", mode="w+")
    write_svf(outfile, svf_bits, c_device, comment=comment)

    outfile.seek(0)
    return outfile



import io
import re
import os.path
import atfu.erase
from atfu.output import Output
from atfu.little_board.jtag_programmer import JtagProgrammer
from atfu.converter.jed2svf import jed2svf
from atfu.converter.svf2xsvf import svf2xsvf


def handler(args):
    output = Output(args)

    if args.erase or args.force:
        erase_result = atfu.erase.perform_erase(args)
        if erase_result != 0:
            exit(erase_result)

    prog = JtagProgrammer("PROGRAM", args.programmer, output.verbosity())

    xsvf_files = _process_input_files(args.device, args.filename, output)
    if prog.upload_all_files(xsvf_files):
        exit(0)
    else:
        exit(1)


def _process_input_files(
    device: str, files: list[io.BufferedReader], output: Output
) -> list[io.BufferedReader]:
    return [_process_input_file(device, result_file, output) for result_file in files]


def _process_input_file(device: str, file: io.BufferedReader, output: Output):
    if re.search(r"[Jj][Ee][Dd]$", file.name):
        return _process_input_svf(output, _process_input_jed(output, device, file))
    elif re.search(r"[^Xx][Ss][Vv][Ff]$", file.name):
        return _process_input_svf(output, file)
    elif re.search(r"[Xx][Ss][Vv][Ff]$", file.name):
        return file
    else:
        output.error(
            "Failure", f"File type for {os.path.basename(file.name)}; Cannot continue"
        )
        exit(80)


def _process_input_jed(
    output: Output, device: str, file: io.BufferedReader
) -> io.BufferedReader:
    return jed2svf(device=device, infile=file, output=output)


def _process_input_svf(output: Output, file: io.BufferedReader) -> io.BufferedReader:
    with file:  # ensure this stays open until we're done so it doesn't get deleted...
        # Bit of weirdness here, the underlying code needs the file to be text...
        inname = file.name

        try:
            infile = open(inname, "r")
        except PermissionError:
            # _Probably_ on Windows, where it doesn't matter anyway...
            infile = file

        return svf2xsvf(output=output, infile=infile)



import argparse
import re

from atfu.converter.jesd3 import JESD3Parser
from atfu.converter.svfeventhandler import SVFEventHandler
from atfu.converter.device import ATF15xxInstr


def read_jed(file):
    jed = file.read()

    # TODO
    #
    # CUPL will emit weird test vectors (with the max vector QV number *after* the
    # first V0001, which _seems_ to be off-spec. The jesd3 code can't handle this.
    #
    # We don't care about test vectors anyway, but we need to at least be able to
    # checksum them, so for now we'll just refuse to process a jed that has vectors
    # in it...
    #
    # Leaving the regexes below in case they're useful late (to be able to just
    # remove the problem vectors **by force**). See also jesd3.py:
    #
    # jed = re.sub(r'\r?\nQV[0-9]+\*', '', jed)
    # jed = re.sub(r'V[0-9+][^\r\n]+\r?\n?\*?', '', jed)

    parser = JESD3Parser(jed)
    parser.parse()
    return parser.fuse, parser.design_spec


class ATFSVFEventHandler(SVFEventHandler):
    def ignored(self, *args, **kwargs):
        pass

    svf_frequency = ignored
    svf_trst = ignored
    svf_state = ignored
    svf_endir = ignored
    svf_enddr = ignored
    svf_hir = ignored
    # svf_sir = ignored
    svf_tir = ignored
    svf_hdr = ignored
    # svf_sdr = ignored
    svf_tdr = ignored
    # svf_runtest = ignored
    svf_piomap = ignored
    svf_pio = ignored

    def __init__(self):
        self.ir = None
        self.erase = False
        self.addr = 0
        self.data = b""
        self.bits = {}

    # def svf_sir(self, tdi, smask, tdo, mask):
    #     self.ir = int.from_bytes(tdi.tobytes(), "little")
    #     if self.ir == ATF15xxInstr.ISC_LATCH_ERASE:
    #         self.erase = True
    #     if self.ir == ATF15xxInstr.ISC_DATA:
    #         self.erase = False

    # def svf_sdr(self, tdi, smask, tdo, mask):
    #     if self.ir == ATF15xxInstr.ISC_ADDRESS:
    #         self.addr = int.from_bytes(tdi.tobytes(), "little")
    #     if (self.ir & ~0x3) == ATF15xxInstr.ISC_DATA:
    #         self.data = tdi

    # def svf_runtest(
    #     self, run_state, run_count, run_clock, min_time, max_time, end_state
    # ):
    #     if not self.erase and self.ir == ATF15xxInstr.ISC_PROGRAM_ERASE:
    #         self.bits[self.addr] = self.data


def write_svf(file, svf_bits, device, *, comment):
    # This code is kind of awful.
    def emit_header():
        for comment_line in comment.splitlines():
            file.write("// {}\n".format(comment_line))
        file.write("TRST ABSENT;\n")
        file.write("ENDIR IDLE;\n")
        file.write("ENDDR IDLE;\n")
        file.write("HDR 0;\n")
        file.write("HIR 0;\n")
        file.write("TDR 0;\n")
        file.write("TIR 0;\n")
        file.write("STATE RESET;\n")

    def emit_check_idcode(idcode):
        file.write("// Check IDCODE\n")
        file.write("SIR 10 TDI ({:03x});\n".format(ATF15xxInstr.IDCODE))
        file.write(
            "SDR 32 TDI (ffffffff)\n\tTDO ({:08x})\n\tMASK (ffffffff);\n".format(idcode)
        )

    def emit_enable():
        file.write("// ISC enable\n")
        file.write("SIR 10 TDI ({:03x});\n".format(ATF15xxInstr.ISC_CONFIG))
        file.write("SDR 10 TDI ({:03x});\n".format(0x1B9))  # magic constant?
        file.write("STATE IDLE;\n")

    def emit_disable():
        file.write("// ISC disable\n")
        file.write("SIR 10 TDI ({:03x});\n".format(ATF15xxInstr.ISC_CONFIG))
        file.write("SDR 10 TDI ({:03x});\n".format(0x000))
        file.write("STATE IDLE;\n")

    def emit_unknown():
        # ATMISP does this for unknown reasons. DR seems to be just BYPASS. Removing this
        # doesn't do anything (and shouldn't do anything, since ATMISP doesn't go through RTI
        # or capture/update DR), but let's keep it for now. Vendor tools wouldn't emit SIR
        # without any reason whatsoever, right? Right??
        file.write("// ISC unknown\n")
        file.write("SIR 10 TDI ({:03x});\n".format(ATF15xxInstr.ISC_UNKNOWN))

    def emit_erase():
        file.write("// ISC erase\n")
        file.write("SIR 10 TDI ({:03x});\n".format(ATF15xxInstr.ISC_LATCH_ERASE))
        file.write("SIR 10 TDI ({:03x});\n".format(ATF15xxInstr.ISC_PROGRAM_ERASE))
        file.write("RUNTEST IDLE 210E-3 SEC;\n")
        emit_unknown()

    def emit_program(address, data):
        file.write("// ISC program word\n")
        file.write("SIR 10 TDI ({:03x});\n".format(ATF15xxInstr.ISC_ADDRESS))
        file.write("SDR 11 TDI ({:03x});\n".format(address))
        file.write(
            "SIR 10 TDI ({:03x});\n".format(ATF15xxInstr.ISC_DATA | (address >> 8))
        )

        # TODO this is a bit of a hack...
        hexout = "{:0{}x}".format(int(data.to01()[::-1], 2), len(data) // 4)
        if (len(data) // 4) & 0x1:
            hexout = "0" + hexout

        file.write("SDR {} TDI ({});\n".format(len(data), hexout))
        file.write("SIR 10 TDI ({:03x});\n".format(ATF15xxInstr.ISC_PROGRAM_ERASE))
        file.write("RUNTEST IDLE 30E-3 SEC;\n")
        emit_unknown()

    def emit_verify(address, data):
        file.write("// ISC verify word\n")
        file.write("SIR 10 TDI ({:03x});\n".format(ATF15xxInstr.ISC_ADDRESS))
        file.write("SDR 11 TDI ({:03x});\n".format(address))
        file.write("SIR 10 TDI ({:03x});\n".format(ATF15xxInstr.ISC_READ))
        file.write("RUNTEST IDLE 20E-3 SEC;\n")
        file.write(
            "SIR 10 TDI ({:03x});\n".format(ATF15xxInstr.ISC_DATA | (address >> 8))
        )

        # TODO this is the same hack as above...
        hexout = "{:0{}x}".format(int(data.to01()[::-1], 2), len(data) // 4)
        if (len(data) // 4) & 0x1:
            hexout = "0" + hexout

        file.write(
            "SDR {} TDI ({})\n\tTDO ({})\n\tMASK ({:0{}x});\n".format(
                len(data),
                hexout,
                hexout,
                (1 << len(data)) - 1,
                len(data) // 4,
            )
        )

    emit_header()
    emit_check_idcode(device.idcode)
    emit_enable()
    emit_erase()
    for svf_row in svf_bits:
        emit_program(svf_row, svf_bits[svf_row])
    for svf_row in svf_bits:
        emit_verify(svf_row, svf_bits[svf_row])
    emit_disable()


class ATFFileType(argparse.FileType):
    def __call__(self, value):
        file = super().__call__(value)
        filename = file.name.lower()
        if not filename.endswith(".jed"):
            raise argparse.ArgumentTypeError("{} is not a JED file".format(filename))
        return file



#!/usr/bin/python3.0

# Copyright 2008, SoftPLC Corporation  http://softplc.com
# Dick Hollenbeck dick@softplc.com


# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, you may find one here:
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.html
# or you may search the http://www.gnu.org website for the version 2 license,
# or you may write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA


# A python program to convert an SVF file to an XSVF file.  There is an
# option to include comments containing the source file line number from the origin
# SVF file before each outputted XSVF statement.

# We deviate from the XSVF spec in that we introduce a new command called
# XWAITSTATE which directly flows from the SVF RUNTEST command.  Unfortunately
# XRUNSTATE was ill conceived and is not used here.  We also add support for the
# three Lattice extensions to SVF: LCOUNT, LDELAY, and LSDR.  The xsvf file
# generated from this program is suitable for use with the xsvf player in
# OpenOCD with my modifications to xsvf.c.

# This program is written for python 3.0, and it is not easy to change this
# back to 2.x.  You may find it easier to use python 3.x even if that means
# building it.

import io
import tempfile
from atfu.output import Output

import re
import sys
import struct


# There are both ---<Lexer>--- and ---<Parser>--- sections to this program

# doCOMMENTs = True  # Save XCOMMENTs in the output xsvf file
doCOMMENTs = False  # Save XCOMMENTs in the output xsvf file

# pick your file encoding
# file_encoding = "ISO-8859-1"
file_encoding = "utf-8"

tokLn = 0
inputFilename = ""

xrepeat = 0  # argument to XREPEAT, gives retry count for masked compares


# -----< Lexer >---------------------------------------------------------------

StateBin = (
    RESET,
    IDLE,
    DRSELECT,
    DRCAPTURE,
    DRSHIFT,
    DREXIT1,
    DRPAUSE,
    DREXIT2,
    DRUPDATE,
    IRSELECT,
    IRCAPTURE,
    IRSHIFT,
    IREXIT1,
    IRPAUSE,
    IREXIT2,
    IRUPDATE,
) = range(16)

# Any integer index into this tuple will be equal to its corresponding StateBin value
StateTxt = (
    "RESET",
    "IDLE",
    "DRSELECT",
    "DRCAPTURE",
    "DRSHIFT",
    "DREXIT1",
    "DRPAUSE",
    "DREXIT2",
    "DRUPDATE",
    "IRSELECT",
    "IRCAPTURE",
    "IRSHIFT",
    "IREXIT1",
    "IRPAUSE",
    "IREXIT2",
    "IRUPDATE",
)


(
    XCOMPLETE,
    XTDOMASK,
    XSIR,
    XSDR,
    XRUNTEST,
    hole0,
    hole1,
    XREPEAT,
    XSDRSIZE,
    XSDRTDO,
    XSETSDRMASKS,
    XSDRINC,
    XSDRB,
    XSDRC,
    XSDRE,
    XSDRTDOB,
    XSDRTDOC,
    XSDRTDOE,
    XSTATE,
    XENDIR,
    XENDDR,
    XSIR2,
    XCOMMENT,
    XWAIT,
    XWAITSTATE,
    LCOUNT,
    LDELAY,
    LSDR,
    XTRST,
) = range(29)

# Note: LCOUNT, LDELAY, and LSDR are Lattice extensions to SVF and provide a way to loop back
# and check a completion status, essentially waiting on a part until it signals that it is done.
# For example below: loop 25 times, each time through the loop do a LDELAY (same as a true RUNTEST)
# and exit loop when LSDR compares match.
"""
LCOUNT	25;
! Step to DRPAUSE give 5 clocks and wait for 1.00e+000 SEC.
LDELAY	DRPAUSE	5 TCK	1.00E-003 SEC;
! Test for the completed status. Match means pass.
! Loop back to LDELAY line if not match and loop count less than 25.
LSDR  1 TDI  (0)
        TDO  (1);
"""

# XTRST is an opcode Xilinx seemed to have missed and it comes from the SVF TRST statement.

LineNumber = 1


def s_ident(scanner, token):
    return ("ident", token.upper(), LineNumber)


def s_hex(scanner, token):
    global LineNumber
    LineNumber = LineNumber + token.count("\n")
    token = "".join(token.split())
    return ("hex", token[1:-1], LineNumber)


def s_int(scanner, token):
    return ("int", int(token), LineNumber)


def s_float(scanner, token):
    return ("float", float(token), LineNumber)


# def s_comment(scanner, token): return ("comment", token, LineNumber)
def s_semicolon(scanner, token):
    return ("semi", token, LineNumber)


def s_nl(scanner, token):
    global LineNumber
    LineNumber = LineNumber + 1
    # print( 'LineNumber=', LineNumber, file=sys.stderr )
    return None


# 2.00E-002


def nextTok(tup):
    """
    Function to read the next token from tup into tokType, tokVal, tokLn (linenumber)
    """
    tokType, tokVal, tokLn = tup.__next__()
    return tokType, tokVal, tokLn


class ParseError(Exception):
    """A class to hold a parsing error message"""

    def __init__(self, linenumber, token, message):
        self.linenumber = linenumber
        self.token = token
        self.message = message

    def __str__(self):
        global inputFilename
        return "Error in file '%s' at line %d near token %s\n %s" % (
            inputFilename,
            self.linenumber,
            repr(self.token),
            self.message,
        )


class MASKSET(object):
    """
    Class MASKSET holds a set of bit vectors, all of which are related, will all
    have the same length, and are associated with one of the seven shiftOps:
    HIR, HDR, TIR, TDR, SIR, SDR, LSDR. One of these holds a mask, smask, tdi, tdo, and a
    size.
    """

    def __init__(self, name):
        self.empty()
        self.name = name

    def empty(self):
        self.mask = bytearray()
        self.smask = bytearray()
        self.tdi = bytearray()
        self.tdo = bytearray()
        self.size = 0

    def syncLengths(self, sawTDI, sawTDO, sawMASK, sawSMASK, newSize):
        """
        Set all the lengths equal in the event some of the masks were
        not seen as part of the last change set.
        """
        if self.size == newSize:
            return

        if newSize == 0:
            self.empty()
            return

        # If an SIR was given without a MASK(), then use a mask of all zeros.
        # this is not consistent with the SVF spec, but it makes sense because
        # it would be odd to be testing an instruction register read out of a
        # tap without giving a mask for it.  Also, lattice seems to agree and is
        # generating SVF files that comply with this philosophy.
        if self.name == "SIR" and not sawMASK:
            self.mask = bytearray(newSize)

        if newSize != len(self.mask):
            self.mask = bytearray(newSize)
            if self.name == "SDR":  # leave mask for HIR,HDR,TIR,TDR,SIR zeros
                for i in range(newSize):
                    self.mask[i] = 1

        if newSize != len(self.tdo):
            self.tdo = bytearray(newSize)

        if newSize != len(self.tdi):
            self.tdi = bytearray(newSize)

        if newSize != len(self.smask):
            self.smask = bytearray(newSize)

        self.size = newSize


# -----</MASKSET>-----


def makeBitArray(hexString, bitCount):
    """
    Converts a packed sequence of hex ascii characters into a bytearray where
    each element in the array holds exactly one bit. Only "bitCount" bits are
    scanned and these must be the least significant bits in the hex number. That
    is, it is legal to have some unused bits in the must significant hex nibble
    of the input "hexString". The string is scanned starting from the backend,
    then just before returning we reverse the array. This way the append()
    method can be used, which I assume is faster than an insert.
    """
    global tokLn
    a = bytearray()
    length = bitCount
    hexString = list(hexString)
    hexString.reverse()
    # print(hexString)
    for c in hexString:
        if length <= 0:
            break
        c = int(c, 16)
        for mask in [1, 2, 4, 8]:
            if length <= 0:
                break
            length = length - 1
            a.append((c & mask) != 0)
    if length > 0:
        raise ParseError(
            tokLn,
            hexString,
            "Insufficient hex characters for given length of %d" % bitCount,
        )
    a.reverse()
    # print(a)
    return a


def makeXSVFbytes(bitarray):
    """
    Make a bytearray which is contains the XSVF bits which will be written
    directly to disk.  The number of bytes needed is calculated from the size
    of the argument bitarray.
    """
    bitCount = len(bitarray)
    byteCount = (bitCount + 7) // 8
    ba = bytearray(byteCount)
    firstBit = (bitCount % 8) - 1
    if firstBit == -1:
        firstBit = 7
    bitNdx = 0
    for byteNdx in range(byteCount):
        mask = 1 << firstBit
        byte = 0
        while mask:
            if bitarray[bitNdx]:
                byte |= mask
            mask = mask >> 1
            bitNdx = bitNdx + 1
        ba[byteNdx] = byte
        firstBit = 7
    return ba


def writeComment(outputFile, shiftOp_linenum, shiftOp):
    """
    Write an XCOMMENT record to outputFile
    """
    comment = "%s @%d\0" % (shiftOp, shiftOp_linenum)  # \0 is terminating nul
    ba = bytearray(1)
    ba[0] = XCOMMENT
    ba += comment.encode()
    outputFile.write(ba)


def combineBitVectors(trailer, meat, header):
    """
    Combine the 3 bit vectors comprizing a transmission.  Since the least
    significant bits are sent first, the header is put onto the list last so
    they are sent first from that least significant position.
    """
    ret = bytearray()
    ret.extend(trailer)
    ret.extend(meat)
    ret.extend(header)
    return ret


def writeRUNTEST(outputFile, run_state, end_state, run_count, min_time, tokenTxt):
    """
    Write the output for the SVF RUNTEST command.
    run_count - the number of clocks
    min_time - the number of seconds
    tokenTxt - either RUNTEST or LDELAY
    """
    # convert from secs to usecs
    min_time = int(min_time * 1000000)

    # the SVF RUNTEST command does NOT map to the XSVF XRUNTEST command.  Check the SVF spec, then
    # read the XSVF command.   They are not the same.  Use an XSVF XWAITSTATE to
    # implement the required behavior of the SVF RUNTEST command.
    if doCOMMENTs:
        writeComment(outputFile, tokLn, tokenTxt)

    if tokenTxt == "RUNTEST":
        obuf = bytearray(11)
        obuf[0] = XWAITSTATE
        obuf[1] = run_state
        obuf[2] = end_state
        struct.pack_into(">i", obuf, 3, run_count)  # big endian 4 byte int to obuf
        struct.pack_into(">i", obuf, 7, min_time)  # big endian 4 byte int to obuf
        outputFile.write(obuf)
    else:  # == 'LDELAY'
        obuf = bytearray(10)
        obuf[0] = LDELAY
        obuf[1] = run_state
        # LDELAY has no end_state
        struct.pack_into(">i", obuf, 2, run_count)  # big endian 4 byte int to obuf
        struct.pack_into(">i", obuf, 6, min_time)  # big endian 4 byte int to obuf
        outputFile.write(obuf)


def svf2xsvf(output: Output, infile: io.BufferedReader) -> io.BufferedReader:
    output.output(2, "Compiling SVF to XSVF...")

    outfile = tempfile.NamedTemporaryFile(
        prefix="atfu-temp-", suffix=".xsvf", mode="w+b"
    )

    hir = MASKSET("HIR")
    hdr = MASKSET("HDR")
    tir = MASKSET("TIR")
    tdr = MASKSET("TDR")
    sir = MASKSET("SIR")
    sdr = MASKSET("SDR")

    expecting_eof = True

    scanner = re.Scanner(
        [
            (r"[a-zA-Z]\w*", s_ident),
            #    (r"[-+]?[0-9]+[.]?[0-9]*([eE][-+]?[0-9]+)?", s_float),
            (r"[-+]?[0-9]+(([.][0-9eE+-]*)|([eE]+[-+]?[0-9]+))", s_float),
            (r"\d+", s_int),
            (r"\(([0-9a-fA-F]|\s)*\)", s_hex),
            (r"(!|//).*$", None),
            (r";", s_semicolon),
            (r"\n", s_nl),
            (r"\s*", None),
        ],
        re.MULTILINE,
    )

    # read all svf file input into string "input"
    #
    # Weirdness: First close it (it'll be binary) and reopen as text
    # (because underlying code needs that...)
    input = infile.read()
    infile.close()

    # Lexer:
    # create a list of tuples containing (tokenType, tokenValue, LineNumber)
    tokens = scanner.scan(input)[0]

    input = None  # allow gc to reclaim memory holding file

    # for tokenType, tokenValue, ln in tokens: print( "line %d: %s" % (ln, tokenType), tokenValue )

    # -----<parser>-----------------------------------------------------------------
    global tokLn
    tokVal = tokType = tokLn = None

    tup = iter(tokens)

    # one of the commands that take the shiftParts after the length, the parse
    # template for all of these commands is identical
    shiftOps = ("SDR", "SIR", "LSDR", "HDR", "HIR", "TDR", "TIR")

    # the order must correspond to shiftOps, this holds the MASKSETS.  'LSDR' shares sdr with 'SDR'
    shiftSets = (sdr, sir, sdr, hdr, hir, tdr, tir)

    # what to expect as parameters to a shiftOp, i.e. after a SDR length or SIR length
    shiftParts = ("TDI", "TDO", "MASK", "SMASK")

    # the set of legal states which can trail the RUNTEST command
    run_state_allowed = ("IRPAUSE", "DRPAUSE", "RESET", "IDLE")

    enddr_state_allowed = ("DRPAUSE", "IDLE")
    endir_state_allowed = ("IRPAUSE", "IDLE")

    trst_mode_allowed = ("ON", "OFF", "Z", "ABSENT")

    enddr_state = IDLE
    endir_state = IDLE

    # frequency = 1.00e006  # HZ;

    # change detection for xsdrsize and xtdomask
    xsdrsize = -1  # the last one sent, send only on change
    xtdomask = bytearray()  # the last one sent, send only on change

    # we use a number of single byte writes for the XSVF command below
    cmdbuf = bytearray(1)

    # Save the XREPEAT setting into the file as first thing.
    obuf = bytearray(2)
    obuf[0] = XREPEAT
    obuf[1] = xrepeat
    outfile.write(obuf)

    try:
        while 1:
            expecting_eof = True
            tokType, tokVal, tokLn = nextTok(tup)
            expecting_eof = False
            # print( tokType, tokVal, tokLn )

            if tokVal in shiftOps:
                shiftOp_linenum = tokLn
                shiftOp = tokVal

                set = shiftSets[shiftOps.index(shiftOp)]

                # set flags false, if we see one later, set that one true later
                sawTDI = sawTDO = sawMASK = sawSMASK = False

                tokType, tokVal, tokLn = nextTok(tup)
                if tokType != "int":
                    raise ParseError(
                        tokLn,
                        tokVal,
                        "Expecting 'int' giving %s length, got '%s'"
                        % (shiftOp, tokType),
                    )
                length = tokVal

                tokType, tokVal, tokLn = nextTok(tup)

                while tokVal != ";":
                    if tokVal not in shiftParts:
                        raise ParseError(
                            tokLn, tokVal, "Expecting TDI, TDO, MASK, SMASK, or ';'"
                        )
                    shiftPart = tokVal

                    tokType, tokVal, tokLn = nextTok(tup)

                    if tokType != "hex":
                        raise ParseError(tokLn, tokVal, "Expecting hex bits")
                    bits = makeBitArray(tokVal, length)

                    if shiftPart == "TDI":
                        sawTDI = True
                        set.tdi = bits

                    elif shiftPart == "TDO":
                        sawTDO = True
                        set.tdo = bits

                    elif shiftPart == "MASK":
                        sawMASK = True
                        set.mask = bits

                    elif shiftPart == "SMASK":
                        sawSMASK = True
                        set.smask = bits

                    tokType, tokVal, tokLn = nextTok(tup)

                set.syncLengths(sawTDI, sawTDO, sawMASK, sawSMASK, length)

                # process all the gathered parameters and generate outputs here
                if shiftOp == "SIR":
                    if doCOMMENTs:
                        writeComment(outfile, shiftOp_linenum, "SIR")

                    tdi = combineBitVectors(tir.tdi, sir.tdi, hir.tdi)
                    if len(tdi) > 255:
                        obuf = bytearray(3)
                        obuf[0] = XSIR2
                        struct.pack_into(">h", obuf, 1, len(tdi))
                    else:
                        obuf = bytearray(2)
                        obuf[0] = XSIR
                        obuf[1] = len(tdi)
                    outfile.write(obuf)
                    obuf = makeXSVFbytes(tdi)
                    outfile.write(obuf)

                elif shiftOp == "SDR":
                    if doCOMMENTs:
                        writeComment(outfile, shiftOp_linenum, shiftOp)

                    if not sawTDO:
                        # pass a zero filled bit vector for the sdr.mask
                        mask = combineBitVectors(
                            tdr.mask, bytearray(sdr.size), hdr.mask
                        )
                        tdi = combineBitVectors(tdr.tdi, sdr.tdi, hdr.tdi)

                        if xsdrsize != len(tdi):
                            xsdrsize = len(tdi)
                            cmdbuf[0] = XSDRSIZE
                            outfile.write(cmdbuf)
                            obuf = bytearray(4)
                            struct.pack_into(
                                ">i", obuf, 0, xsdrsize
                            )  # big endian 4 byte int to obuf
                            outfile.write(obuf)

                        if xtdomask != mask:
                            xtdomask = mask
                            cmdbuf[0] = XTDOMASK
                            outfile.write(cmdbuf)
                            obuf = makeXSVFbytes(mask)
                            outfile.write(obuf)

                        cmdbuf[0] = XSDR
                        outfile.write(cmdbuf)
                        obuf = makeXSVFbytes(tdi)
                        outfile.write(obuf)

                    else:
                        mask = combineBitVectors(tdr.mask, sdr.mask, hdr.mask)
                        tdi = combineBitVectors(tdr.tdi, sdr.tdi, hdr.tdi)
                        tdo = combineBitVectors(tdr.tdo, sdr.tdo, hdr.tdo)

                        if xsdrsize != len(tdi):
                            xsdrsize = len(tdi)
                            cmdbuf[0] = XSDRSIZE
                            outfile.write(cmdbuf)
                            obuf = bytearray(4)
                            struct.pack_into(
                                ">i", obuf, 0, xsdrsize
                            )  # big endian 4 byte int to obuf
                            outfile.write(obuf)

                        if xtdomask != mask:
                            xtdomask = mask
                            cmdbuf[0] = XTDOMASK
                            outfile.write(cmdbuf)
                            obuf = makeXSVFbytes(mask)
                            outfile.write(obuf)

                        cmdbuf[0] = XSDRTDO
                        outfile.write(cmdbuf)
                        obuf = makeXSVFbytes(tdi)
                        outfile.write(obuf)
                        obuf = makeXSVFbytes(tdo)
                        outfile.write(obuf)
                        # print( "len(tdo)=", len(tdo), "len(tdr.tdo)=", len(tdr.tdo), "len(sdr.tdo)=", len(sdr.tdo), "len(hdr.tdo)=", len(hdr.tdo) )

                elif shiftOp == "LSDR":
                    if doCOMMENTs:
                        writeComment(outfile, shiftOp_linenum, shiftOp)

                    mask = combineBitVectors(tdr.mask, sdr.mask, hdr.mask)
                    tdi = combineBitVectors(tdr.tdi, sdr.tdi, hdr.tdi)
                    tdo = combineBitVectors(tdr.tdo, sdr.tdo, hdr.tdo)

                    if xsdrsize != len(tdi):
                        xsdrsize = len(tdi)
                        cmdbuf[0] = XSDRSIZE
                        outfile.write(cmdbuf)
                        obuf = bytearray(4)
                        struct.pack_into(
                            ">i", obuf, 0, xsdrsize
                        )  # big endian 4 byte int to obuf
                        outfile.write(obuf)

                    if xtdomask != mask:
                        xtdomask = mask
                        cmdbuf[0] = XTDOMASK
                        outfile.write(cmdbuf)
                        obuf = makeXSVFbytes(mask)
                        outfile.write(obuf)

                    cmdbuf[0] = LSDR
                    outfile.write(cmdbuf)
                    obuf = makeXSVFbytes(tdi)
                    outfile.write(obuf)
                    obuf = makeXSVFbytes(tdo)
                    outfile.write(obuf)
                    # print( "len(tdo)=", len(tdo), "len(tdr.tdo)=", len(tdr.tdo), "len(sdr.tdo)=", len(sdr.tdo), "len(hdr.tdo)=", len(hdr.tdo) )

            elif tokVal == "RUNTEST" or tokVal == "LDELAY":
                # e.g. from lattice tools:
                # "RUNTEST	IDLE  	5 TCK	1.00E-003 SEC;"
                saveTok = tokVal
                tokType, tokVal, tokLn = nextTok(tup)
                min_time = 0
                run_count = 0
                # max_time = 600  # ten minutes
                if tokVal in run_state_allowed:
                    run_state = StateTxt.index(tokVal)
                    end_state = run_state  # bottom of page 17 of SVF spec
                    tokType, tokVal, tokLn = nextTok(tup)
                if tokType != "int" and tokType != "float":
                    raise ParseError(
                        tokLn,
                        tokVal,
                        "Expecting 'int' or 'float' after RUNTEST [run_state]",
                    )
                timeval = tokVal
                tokType, tokVal, tokLn = nextTok(tup)
                if tokVal != "TCK" and tokVal != "SEC" and tokVal != "SCK":
                    raise ParseError(
                        tokLn,
                        tokVal,
                        "Expecting 'TCK' or 'SEC' or 'SCK' after RUNTEST [run_state] (run_count|min_time)",
                    )
                if tokVal == "TCK" or tokVal == "SCK":
                    run_count = int(timeval)
                else:
                    min_time = timeval
                tokType, tokVal, tokLn = nextTok(tup)
                if tokType == "int" or tokType == "float":
                    min_time = tokVal
                    tokType, tokVal, tokLn = nextTok(tup)
                    if tokVal != "SEC":
                        raise ParseError(
                            tokLn,
                            tokVal,
                            "Expecting 'SEC' after RUNTEST [run_state] run_count min_time",
                        )
                    tokType, tokVal, tokLn = nextTok(tup)
                if tokVal == "MAXIMUM":
                    tokType, tokVal, tokLn = nextTok(tup)
                    if tokType != "int" and tokType != "float":
                        raise ParseError(
                            tokLn,
                            tokVal,
                            "Expecting 'max_time' after RUNTEST [run_state] min_time SEC MAXIMUM",
                        )
                    # max_time = tokVal
                    tokType, tokVal, tokLn = nextTok(tup)
                    if tokVal != "SEC":
                        raise ParseError(
                            tokLn,
                            tokVal,
                            "Expecting 'max_time' after RUNTEST [run_state] min_time SEC MAXIMUM max_time",
                        )
                    tokType, tokVal, tokLn = nextTok(tup)
                if tokVal == "ENDSTATE":
                    tokType, tokVal, tokLn = nextTok(tup)
                    if tokVal not in run_state_allowed:
                        raise ParseError(
                            tokLn,
                            tokVal,
                            "Expecting 'run_state' after RUNTEST .... ENDSTATE",
                        )
                    end_state = StateTxt.index(tokVal)
                    tokType, tokVal, tokLn = nextTok(tup)
                if tokVal != ";":
                    raise ParseError(tokLn, tokVal, "Expecting ';' after RUNTEST ....")
                    # print( "run_count=", run_count, "min_time=", min_time,
                    # "max_time=", max_time, "run_state=", State[run_state], "end_state=", State[end_state] )
                    print(f"{tokVal} == {tokType}")
                    writeRUNTEST(
                        outfile, run_state, end_state, run_count, min_time, saveTok
                    )

            elif tokVal == "LCOUNT":
                tokType, tokVal, tokLn = nextTok(tup)
                if tokType != "int":
                    raise ParseError(
                        tokLn, tokVal, "Expecting integer 'count' after LCOUNT"
                    )
                loopCount = tokVal
                tokType, tokVal, tokLn = nextTok(tup)
                if tokVal != ";":
                    raise ParseError(tokLn, tokVal, "Expecting ';' after LCOUNT count")
                if doCOMMENTs:
                    writeComment(outfile, tokLn, "LCOUNT")
                obuf = bytearray(5)
                obuf[0] = LCOUNT
                struct.pack_into(
                    ">i", obuf, 1, loopCount
                )  # big endian 4 byte int to obuf
                outfile.write(obuf)

            elif tokVal == "ENDDR":
                tokType, tokVal, tokLn = nextTok(tup)
                if tokVal not in enddr_state_allowed:
                    raise ParseError(
                        tokLn,
                        tokVal,
                        "Expecting 'stable_state' after ENDDR. (one of: DRPAUSE, IDLE)",
                    )
                enddr_state = StateTxt.index(tokVal)
                tokType, tokVal, tokLn = nextTok(tup)
                if tokVal != ";":
                    raise ParseError(
                        tokLn, tokVal, "Expecting ';' after ENDDR stable_state"
                    )
                if doCOMMENTs:
                    writeComment(outfile, tokLn, "ENDDR")
                obuf = bytearray(2)
                obuf[0] = XENDDR
                # Page 10 of the March 1999 SVF spec shows that RESET is also allowed here.
                # Yet the XSVF spec has no provision for that, and uses a non-standard, i.e.
                # boolean argument to XENDDR which only handles two of the 3 intended states.
                obuf[1] = 1 if enddr_state == DRPAUSE else 0
                outfile.write(obuf)

            elif tokVal == "ENDIR":
                tokType, tokVal, tokLn = nextTok(tup)
                if tokVal not in endir_state_allowed:
                    raise ParseError(
                        tokLn,
                        tokVal,
                        "Expecting 'stable_state' after ENDIR. (one of: IRPAUSE, IDLE)",
                    )
                endir_state = StateTxt.index(tokVal)
                tokType, tokVal, tokLn = nextTok(tup)
                if tokVal != ";":
                    raise ParseError(
                        tokLn, tokVal, "Expecting ';' after ENDIR stable_state"
                    )
                if doCOMMENTs:
                    writeComment(outfile, tokLn, "ENDIR")
                obuf = bytearray(2)
                obuf[0] = XENDIR
                # Page 10 of the March 1999 SVF spec shows that RESET is also allowed here.
                # Yet the XSVF spec has no provision for that, and uses a non-standard, i.e.
                # boolean argument to XENDDR which only handles two of the 3 intended states.
                obuf[1] = 1 if endir_state == IRPAUSE else 0
                outfile.write(obuf)

            elif tokVal == "STATE":
                tokType, tokVal, tokLn = nextTok(tup)
                ln = tokLn
                while tokVal != ";":
                    if tokVal not in StateTxt:
                        raise ParseError(
                            tokLn, tokVal, "Expecting 'stable_state' after STATE"
                        )
                    stable_state = StateTxt.index(tokVal)

                    if doCOMMENTs and ln != -1:
                        writeComment(outfile, ln, "STATE")
                        ln = -1  # save comment only once

                    obuf = bytearray(2)
                    obuf[0] = XSTATE
                    obuf[1] = stable_state
                    outfile.write(obuf)
                    tokType, tokVal, tokLn = nextTok(tup)

            elif tokVal == "FREQUENCY":
                tokType, tokVal, tokLn = nextTok(tup)
                if tokVal != ";":
                    if tokType != "int" and tokType != "float":
                        raise ParseError(
                            tokLn, tokVal, "Expecting 'cycles HZ' after FREQUENCY"
                        )
                    # frequency = tokVal
                    tokType, tokVal, tokLn = nextTok(tup)
                    if tokVal != "HZ":
                        raise ParseError(
                            tokLn, tokVal, "Expecting 'HZ' after FREQUENCY cycles"
                        )
                    tokType, tokVal, tokLn = nextTok(tup)
                    if tokVal != ";":
                        raise ParseError(
                            tokLn, tokVal, "Expecting ';' after FREQUENCY cycles HZ"
                        )

            elif tokVal == "TRST":
                tokType, tokVal, tokLn = nextTok(tup)
                if tokVal not in trst_mode_allowed:
                    raise ParseError(
                        tokLn, tokVal, "Expecting 'ON|OFF|Z|ABSENT' after TRST"
                    )
                trst_mode = tokVal
                tokType, tokVal, tokLn = nextTok(tup)
                if tokVal != ";":
                    raise ParseError(
                        tokLn, tokVal, "Expecting ';' after TRST trst_mode"
                    )
                if doCOMMENTs:
                    writeComment(outfile, tokLn, "TRST %s" % trst_mode)
                obuf = bytearray(2)
                obuf[0] = XTRST
                obuf[1] = trst_mode_allowed.index(
                    trst_mode
                )  # use the index as the binary argument to XTRST opcode
                outfile.write(obuf)

            else:
                raise ParseError(tokLn, tokVal, "Unknown token '%s'" % tokVal)

    except StopIteration:
        if not expecting_eof:
            output.error("Failure", "SVF: Unexpected End of File at line ", tokLn)

    except ParseError as pe:
        output.error("Failure", f"SVF: {pe}")

    finally:
        # print( "closing file" )
        cmdbuf[0] = XCOMPLETE
        outfile.write(cmdbuf)

    outfile.seek(0)
    return outfile



# Be careful modifying this file: line ranges from it are included in docs/jtag/as.rst.

import enum
from bitarray import bitarray


__all__ = ["ATF15xxInstr", "ATF1502ASDevice", "ATF1504ASDevice", "ATF1508ASDevice"]


class ATF15xxInstr(enum.IntEnum):
    EXTEST = 0x000
    SAMPLE = 0x055
    IDCODE = 0x059
    ISC_READ_UES = 0x270
    ISC_CONFIG = 0x280
    ISC_READ = 0x28C
    ISC_DATA = 0x290
    ISC_PROGRAM_ERASE = 0x29E
    ISC_ADDRESS = 0x2A1
    ISC_LATCH_ERASE = 0x2B3
    ISC_UNKNOWN = 0x2BF
    BYPASS = 0x3FF


class ATF15xxDevice:
    idcode = None  # int

    fuse_count = None  # int
    data_width = None  # dict(range/tuple,range)

    @classmethod
    def word_size(cls, svf_row):
        for svf_rows, svf_cols in cls.data_width.items():
            if svf_row in svf_rows:
                return len(svf_cols)
        assert False

    @staticmethod
    def jed_to_svf_coords(jed_index):
        raise NotImplementedError

    @classmethod
    def jed_to_svf(cls, jed_bits):
        svf_bits = {}
        for jed_index, jed_bit in enumerate(jed_bits):
            svf_index = cls.jed_to_svf_coords(jed_index)
            if svf_index is None:
                continue
            svf_row, svf_col = svf_index
            if svf_row not in svf_bits:
                svf_bits[svf_row] = bitarray(cls.word_size(svf_row))
                svf_bits[svf_row].setall(1)
            svf_bits[svf_row][svf_col] = jed_bit
        return svf_bits

    @staticmethod
    def svf_to_jed_coords(svf_row, svf_col):
        raise NotImplementedError

    @classmethod
    def svf_to_jed(cls, svf_bits):
        jed_bits = bitarray(cls.fuse_count)
        jed_bits.setall(0)
        for svf_row, svf_word in svf_bits.items():
            for svf_col, svf_bit in enumerate(svf_word):
                jed_index = cls.svf_to_jed_coords(svf_row, svf_col)
                if jed_index is None:
                    continue
                jed_bits[jed_index] = svf_bit
        return jed_bits


class ATF1502ASDevice(ATF15xxDevice):
    idcode = 0x0150203F

    fuse_count = 16808
    data_width = {
        range(0, 108): range(86),
        range(128, 229): range(86),
        (256,): range(32),
        (512,): range(4),
        (768,): range(16),
    }

    @staticmethod
    def jed_to_svf_coords(jed_index):
        if jed_index in range(0, 7680):
            return 12 + (jed_index - 0) % 96, 79 - (jed_index - 0) // 96
        if jed_index in range(7680, 15360):
            return 128 + (jed_index - 7680) % 96, 79 - (jed_index - 7680) // 96
        if jed_index in range(15360, 16320):
            return 0 + (jed_index - 15360) // 80, 79 - (jed_index - 15360) % 80
        if jed_index in range(16320, 16720):
            return 224 + (jed_index - 16320) % 5, 79 - (jed_index - 16320) // 5
        if jed_index in range(16720, 16750):
            return 224 + (jed_index - 16320) % 5, 85 - (jed_index - 16320) // 5 + 80
        if jed_index in range(16750, 16782):
            return 256, 31 - (jed_index - 16750)
        if jed_index in range(16782, 16786):
            return 512, 3 - (jed_index - 16782)
        if jed_index in range(16786, 16802):
            return 768, 15 - (jed_index - 16786)
        if jed_index in range(16802, 16808):
            return  # reserved
        assert False

    @staticmethod
    def svf_to_jed_coords(svf_row, svf_col):
        if svf_row in range(0, 12):
            if svf_col in range(0, 80):
                return 15360 + (svf_row - 0) * 80 + (79 - svf_col)
            else:
                return  # always 1
        if svf_row in range(12, 108):
            if svf_col in range(0, 80):
                return 0 + (svf_row - 12) + (79 - svf_col) * 96
            else:
                return  # always 1
        if svf_row in range(128, 224):
            if svf_col in range(0, 80):
                return 7680 + (svf_row - 128) + (79 - svf_col) * 96
            else:
                return  # always 1
        if svf_row in range(224, 229):
            if svf_col in range(0, 80):
                return 16320 + (svf_row - 224) + (79 - svf_col) * 5
            else:
                return 16720 + (svf_row - 224) + (85 - svf_col) * 5
        if svf_row == 256:
            return 16750 + (31 - svf_col)
        if svf_row == 512:
            return 16782 + (3 - svf_col)
        if svf_row == 768:
            return 16786 + (15 - svf_col)
        assert False


class ATF1504ASDevice(ATF15xxDevice):
    idcode = 0x0150403F

    fuse_count = 34192
    data_width = {
        range(0, 108): range(166),
        range(128, 233): range(166),
        (256,): range(32),
        (512,): range(4),
        (768,): range(16),
    }

    @staticmethod
    def jed_to_svf_coords(jed_index):
        if jed_index in range(0, 15360):
            return 12 + (jed_index - 0) % 96, 165 - (jed_index - 0) // 96
        if jed_index in range(15360, 30720):
            return 128 + (jed_index - 15360) % 96, 165 - (jed_index - 15360) // 96
        if jed_index in range(30720, 32640):
            return 0 + (jed_index - 30720) // 160, 165 - (jed_index - 30720) % 160
        if jed_index in range(32640, 34134):
            return 224 + (jed_index - 32640) % 9, 165 - (jed_index - 32640) // 9
        if jed_index in range(34134, 34166):
            return 256, 31 - (jed_index - 34134)
        if jed_index in range(34166, 34170):
            return 512, 3 - (jed_index - 34166)
        if jed_index in range(34170, 34186):
            return 768, 15 - (jed_index - 34170)
        if jed_index in range(34186, 34192):
            return  # reserved
        assert False

    @staticmethod
    def svf_to_jed_coords(svf_row, svf_col):
        if svf_row in range(0, 12):
            if svf_col in range(6, 166):
                return 30720 + (svf_row - 0) * 160 + (165 - svf_col)
            else:
                return  # always 1
        if svf_row in range(12, 108):
            if svf_col in range(6, 166):
                return 0 + (svf_row - 12) + (165 - svf_col) * 96
            else:
                return  # always 1
        if svf_row in range(128, 224):
            if svf_col in range(6, 166):
                return 15360 + (svf_row - 128) + (165 - svf_col) * 96
            else:
                return  # always 1
        if svf_row in range(224, 233):
            if svf_col in range(0, 166):
                return 32640 + (svf_row - 224) + (165 - svf_col) * 9
            else:
                return  # always 1
        if svf_row == 256:
            return 34134 + (31 - svf_col)
        if svf_row == 512:
            return 34166 + (3 - svf_col)
        if svf_row == 768:
            return 34170 + (15 - svf_col)
        assert False


class ATF1508ASDevice(ATF15xxDevice):
    idcode = 0x0150803F

    fuse_count = 74136
    data_width = {
        range(0, 108): range(326),
        range(128, 251): range(326),
        (256,): range(32),
        (512,): range(4),
        (768,): range(16),
    }

    @staticmethod
    def jed_to_svf_coords(jed_index):
        if jed_index in range(0, 30720):
            return 12 + (jed_index - 0) % 96, 325 - (jed_index - 0) // 96
        if jed_index in range(30720, 61440):
            return 128 + (jed_index - 30720) % 96, 325 - (jed_index - 30720) // 96
        if jed_index in range(61440, 65280):
            return 0 + (jed_index - 61440) // 320, 325 - (jed_index - 61440) % 320
        if jed_index in range(65280, 74082):
            return 224 + (jed_index - 65280) % 27, 325 - (jed_index - 65280) // 27
        if jed_index in range(74082, 74114):
            return 256, 31 - (jed_index - 74082)
        if jed_index in range(74114, 74118):
            return 512, 3 - (jed_index - 74114)
        if jed_index in range(74118, 74134):
            return 768, 15 - (jed_index - 74118)
        if jed_index in range(74134, 74136):
            return  # reserved
        assert False

    @staticmethod
    def svf_to_jed_coords(svf_row, svf_col):
        if svf_row in range(0, 12):
            if svf_col in range(6, 326):
                return 61440 + (svf_row - 0) * 320 + (325 - svf_col)
            else:
                return  # always 1
        if svf_row in range(12, 108):
            if svf_col in range(6, 326):
                return 0 + (svf_row - 12) + (325 - svf_col) * 96
            else:
                return  # always 1
        if svf_row in range(128, 224):
            if svf_col in range(6, 326):
                return 30720 + (svf_row - 128) + (325 - svf_col) * 96
            else:
                return  # always 1
        if svf_row in range(224, 251):
            if svf_col in range(0, 326):
                return 65280 + (svf_row - 224) + (325 - svf_col) * 27
            else:
                return  # always 1
        if svf_row == 256:
            return 74082 + (31 - svf_col)
        if svf_row == 512:
            return 74114 + (3 - svf_col)
        if svf_row == 768:
            return 74118 + (15 - svf_col)
        assert False
